/*
 * Copyright (c) 2015. Rick Hightower, Geoff Chandler
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * QBit - The Microservice lib for Java : JSON, WebSocket, REST. Be The Web!
 */

package io.advantageous.qbit.http.client;

import io.advantageous.qbit.GlobalConstants;
import io.advantageous.qbit.QBit;
import io.advantageous.qbit.config.PropertyResolver;

import java.util.Properties;


/**
 * This allows one to construct an http client which attaches to a remote server.
 *
 * @author rhightower
 *         created by rhightower on 11/13/14.
 */
public class HttpClientBuilder {


    public static final String QBIT_HTTP_CLIENT_BUILDER = "qbit.http.client.builder.";

    private String host = "localhost";
    private int port = 8080;
    private int poolSize = 20;
    private int timeOutInMilliseconds = 3_000;
    private boolean autoFlush = true;
    private boolean keepAlive = true;
    private boolean pipeline = true;
    private int flushInterval = 500;


    public HttpClientBuilder(PropertyResolver propertyResolver) {
        this.autoFlush = propertyResolver.getBooleanProperty("autoFlush", true);
        this.host = propertyResolver.getStringProperty("host", "localhost");
        this.port = propertyResolver.getIntegerProperty("port", 8080);
        this.poolSize = propertyResolver.getIntegerProperty("poolSize", 1);
        this.keepAlive = propertyResolver.getBooleanProperty("keepAlive", true);
        this.pipeline = propertyResolver.getBooleanProperty("pipeline", true);
        this.timeOutInMilliseconds = propertyResolver.getIntegerProperty("timeOutInMilliseconds", 3000);
        this.flushInterval = propertyResolver.getIntegerProperty("flushInterval", 500);
    }


    public HttpClientBuilder() {
        this(PropertyResolver.createSystemPropertyResolver(QBIT_HTTP_CLIENT_BUILDER));
    }


    public HttpClientBuilder(final Properties properties) {
        this(PropertyResolver.createPropertiesPropertyResolver(
                QBIT_HTTP_CLIENT_BUILDER, properties));
    }

    public static HttpClientBuilder httpClientBuilder() {

        return new HttpClientBuilder();
    }

    public int getFlushInterval() {
        return flushInterval;
    }

    public HttpClientBuilder setFlushInterval(int flushInterval) {
        this.flushInterval = flushInterval;
        return this;
    }

    public boolean isKeepAlive() {
        return keepAlive;
    }

    public HttpClientBuilder setKeepAlive(boolean keepAlive) {
        this.keepAlive = keepAlive;
        return this;
    }

    public boolean isPipeline() {
        return pipeline;
    }

    public HttpClientBuilder setPipeline(boolean pipeline) {
        this.pipeline = pipeline;
        return this;
    }

    public String getHost() {
        return host;
    }

    public HttpClientBuilder setHost(String host) {
        this.host = host;
        return this;
    }

    public int getPort() {
        return port;
    }

    public HttpClientBuilder setPort(int port) {
        this.port = port;
        return this;
    }



    public int getPoolSize() {
        return poolSize;
    }

    public HttpClientBuilder setPoolSize(int poolSize) {
        this.poolSize = poolSize;
        return this;
    }

    public int getTimeOutInMilliseconds() {
        return timeOutInMilliseconds;
    }

    public HttpClientBuilder setTimeOutInMilliseconds(int timeOutInMilliseconds) {
        this.timeOutInMilliseconds = timeOutInMilliseconds;
        return this;
    }

    public boolean isAutoFlush() {
        return autoFlush;
    }

    public HttpClientBuilder setAutoFlush(boolean autoFlush) {
        this.autoFlush = autoFlush;
        return this;
    }

    public HttpClient build() {
        //noinspection UnnecessaryLocalVariable
        @SuppressWarnings("UnnecessaryLocalVariable") final HttpClient httpClient = QBit.factory().createHttpClient(
                this.getHost(),
                this.getPort(),
                this.getTimeOutInMilliseconds(),
                this.getPoolSize(),
                this.isAutoFlush(),
                this.getFlushInterval(),
                this.isKeepAlive(),
                this.isPipeline());

        return httpClient;
    }


    public HttpClient buildAndStart() {
        HttpClient client = build();
        client.start();
        return client;
    }
}
